/*
 *  SMC1 raw console serial I/O.
 *
 *  This driver is an example of `POLLING' or `INTERRUPT' I/O.
 *
 *  To run with interrupt-driven I/O, ensure m360_smc1_interrupt
 *  is set before calling the initialization routine.
 *
 *  Author:
 *    W. Eric Norum
 *    Saskatchewan Accelerator Laboratory
 *    University of Saskatchewan
 *    Saskatoon, Saskatchewan, CANADA
 *    eric@skatter.usask.ca
 *
 *  COPYRIGHT (c) 1989-1998.
 *  On-Line Applications Research Corporation (OAR).
 *  Copyright assigned to U.S. Government, 1994.
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *
 *  http://www.OARcorp.com/rtems/license.html.
 *
 *  $Id: console.c,v 1.11 1998/02/19 23:02:02 joel Exp $
 */

#include <termios.h>
#include <bsp.h>
#include <rtems/libio.h>

/*
 * Interrupt-driven callback
 */
static int mcf_uart_interrupt = 1;
static void *mcf_uartttyp;

volatile uartstruct* uart; /* this stores the uart channel being used */

/*
 * Device-specific routines
 */
static rtems_isr
mcf_uartInterruptHandler (rtems_vector_number v)
{
	char ch;
	
	/*
	 * Buffer received?
	 */
	if (uart->read.uisr & MCF5206_UART_UISR_RXRDY) {
		while ((uart->read.usr & MCF5206_UART_USR_RXRDY) != 0) {
			ch = uart->read.urbuf;
			rtems_termios_enqueue_raw_characters( mcf_uartttyp, &ch, 1 );
		}
	}

	/*
	 * Buffer transmitted?
	 */
	if (uart->read.uisr & MCF5206_UART_UISR_TXRDY) {
		rtems_termios_dequeue_characters( mcf_uartttyp, 1 ); /* only one char at a time */
	}
}

static void
mcf_uartInitialize (void)
{

	uart = &sim->uart1;
	
	uart->write.ucr = MCF5206_UART_UCR_RESET_MR |
	                  MCF5206_UART_UCR_RESET_TX |
	                  MCF5206_UART_UCR_RESET_RX;
			  
	sim->icr12 = MCF5206_SIM_ICR_IL(UART1_IRQ_LEVEL) |
	             MCF5206_SIM_ICR_IP(UART1_IRQ_PRIORITY);
		     
	uart->write.uivr = UART1_IRQ_VECTOR;
	
	uart->write.ubg1 =  0; /* 9600 baud */
	uart->write.ubg2 = 0x28;
	
	uart->write.ucsr = 0xDD; /* internal baud rate generator */
	uart->write.umr  = MCF5206_UART_UMR1_PM_NONE | /* UMR1 = 8-bit, no parity */
	                   MCF5206_UART_UMR1_BC_8;
	uart->write.umr  = MCF5206_UART_UMR2_CM_NORMAL  |
			   MCF5206_UART_UMR2_STOP_BITS_1;
			 
	uart->write.ucr = MCF5206_UART_UCR_TX_ENABLED |
	                  MCF5206_UART_UCR_RX_ENABLED;

	if (mcf_uart_interrupt) {
	rtems_isr_entry old_handler;
	rtems_status_code sc;

	sc = rtems_interrupt_catch (mcf_uartInterruptHandler,
						UART1_IRQ_VECTOR,
						&old_handler);
	uart->write.uimr = MCF5206_UART_UIMR_FFULL | /* Enable SMC1 TX and RX interrupts */
	                   MCF5206_UART_UIMR_TXRDY;
	sim->imr = sim->imr & (~MCF5206_SIM_IMR_UART1);	/* Enable UART1 interrupts */
	}
}

static int
mcf_uartRead (int minor)
{
	if ((uart->read.usr & MCF5206_UART_USR_RXRDY) == 0)
		return -1;
		
	return uart->read.urbuf;
}

/*
 * Device-dependent write routine
 * Interrupt-driven devices:
 *	Begin transmission of as many characters as possible (minimum is 1).
 * Polling devices:
 *	Transmit all characters.
 */
static int
mcf_uartInterruptWrite (int minor, const char *buf, int len)
{
	if ( len > 0 )
		uart->write.utbuf = *buf;
	return 0;
}

static int
mcf_uartPollWrite (int minor, const char *buf, int len)
{
	while (len--) {
		static char txBuf;
		while ( (uart->read.usr & MCF5206_UART_USR_TXRDY) == 0 )
			continue;
		txBuf = *buf++;
		uart->write.utbuf = txBuf;
	}
	return 0;
}

/*
 ***************
 * BOILERPLATE *
 ***************
 */

/*
 * Reserve resources consumed by this driver
 */
void console_reserve_resources(
  rtems_configuration_table *configuration
)
{
	rtems_termios_reserve_resources (configuration, 1);
}

/*
 * Initialize and register the device
 */
rtems_device_driver console_initialize(
  rtems_device_major_number  major,
  rtems_device_minor_number  minor,
  void                      *arg
)
{
	rtems_status_code status;

	/*
	 * Set up TERMIOS
	 */
	rtems_termios_initialize ();

	/*
	 * Do device-specific initialization
	 */
	mcf_uartInitialize ();

	/*
	 * Register the device
	 */
	status = rtems_io_register_name ("/dev/console", major, 0);
	if (status != RTEMS_SUCCESSFUL)
		rtems_fatal_error_occurred (status);
	return RTEMS_SUCCESSFUL;
}

/*
 * Open the device
 */
rtems_device_driver console_open(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
	rtems_status_code sc;

	if (mcf_uart_interrupt) {
		rtems_libio_open_close_args_t *args = arg;

		sc = rtems_termios_open (major, minor, arg,
					NULL,
					NULL,
					NULL,
					mcf_uartInterruptWrite,
					1);
		mcf_uartttyp = args->iop->data1;
	}
	else {
		sc = rtems_termios_open (major, minor, arg,
					NULL,
					NULL,
					mcf_uartRead,
					mcf_uartPollWrite,
					0);
	}
	return sc;
}
 
/*
 * Close the device
 */
rtems_device_driver console_close(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
	return rtems_termios_close (arg);
}

/*
 * Read from the device
 */
rtems_device_driver console_read(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
	return rtems_termios_read (arg);
}

/*
 * Write to the device
 */
rtems_device_driver console_write(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
	return rtems_termios_write (arg);
}

/*
 * Handle ioctl request.
 * Should set hardware line speed, bits/char, etc.
 */
rtems_device_driver console_control(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{ 
	return rtems_termios_ioctl (arg);
}
